require 'spec_helper'
require 'emcee'
require 'emcee/file_cache'

class Emcee::CacheBase
  def log(message)
  end
end

describe Emcee::FileCache do
  before do
    @key = { 'foo' => { 'bar' => 'baz' }, 'xyzzy' => [ 1, 2, 3, [ 'a', 'b', 'c' ] ] }
  end

  describe 'when initializing' do
    it 'accepts a key' do
      lambda { Emcee::FileCache.new(@key) }.should_not raise_error(ArgumentError)
    end

    it 'requires a key' do
      lambda { Emcee::FileCache.new }.should raise_error(ArgumentError)
    end

    it 'makes the key available after initialization' do
      Emcee::FileCache.new(@key).key.should == @key
    end
  end

  describe 'when accessing data' do
    before do
      @cache = Emcee::FileCache.new(@key)
    end

    # clean up any created cache files
    after do
      File.unlink(@cache.cache_path) if File.exists?(@cache.cache_path)
    end

    describe 'and data has not been seen before' do

      it 'runs the data block' do
        lambda { @cache.data do raise "flunk" end }.should raise_error(RuntimeError)
      end

      it 'writes the data to the cache path as YAML' do
        @cache.data do [ 'x', 'y' ] end
        YAML.load(File.read(@cache.cache_path)).should == [ 'x', 'y' ]
      end

      it 'returns the data' do
        @cache.data do [ 'x', 'y' ] end.should == [ 'x', 'y' ]
      end
    end

    describe 'and data has been seen before' do
      before do
        @cache.data do [ 'x', 'y' ] end
      end

      describe 'and cached data is stale' do
        before do
          @cache.stubs(:fresh?).returns(false)
        end

        it 'runs the data block' do
          lambda { @cache.data do raise "flunk" end }.should raise_error(RuntimeError)
        end

        it 're-writes the data to the cache path as YAML' do
          @cache.data do [ 'x', 'y' ] end
          YAML.load(File.read(@cache.cache_path)).should == [ 'x', 'y' ]
          (Time.now - File.mtime(@cache.cache_path)).should < 2 # presume it takes less than a couple of seconds to run the test
        end

        it 'returns the data' do
          @cache.data do [ 'x', 'y' ] end.should == [ 'x', 'y' ]
        end
      end

      describe 'and cached data is not stale' do
        before do
          @cache.stubs(:fresh?).returns(true)
        end

        it 'does not store the data' do
          @cache.expects(:save).never
          @cache.data do [ 'x', 'y' ] end
        end

        it 'does not run the data block' do
          lambda { @cache.data do raise "flunk" end }.should_not raise_error(RuntimeError)
        end

        it 'returns the data' do
          @cache.data do [ 'x', 'y' ] end.should == [ 'x', 'y' ]
        end
      end
    end
  end

  describe 'when manually expiring the cache' do
    before do
      @cache = Emcee::FileCache.new(@key)
    end

    it 'deletes the cache file if it exists' do
      @cache.data do [ 'x', 'y' ] end
      File.exists?(@cache.cache_path).should be_true
      @cache.expire!
      File.exists?(@cache.cache_path).should be_false
    end

    it 'does nothing if no cache file exists' do
      File.expects(:delete).never
      @cache.expire!
    end
  end
end
