require 'spec_helper'
require 'emcee/result'

describe Emcee::Result do
  before do
    @contents = { :output => { 'foo' => 'bar' } }
    @node_name = 'somenode'
    @status_code = 0
    @status_message = 'totally ruined.'
    @results = { :sender => @node_name, :data => @contents, :statuscode => @status_code, :statusmsg => @status_message }
    @result = stub(:results => @results)
  end

  describe 'when initializing' do
    it 'takes an MCollective result object' do
      lambda { Emcee::Result.new(@result) }.should_not raise_error(ArgumentError)
    end

    it 'requires an MCollective result object' do
      lambda { Emcee::Result.new }.should raise_error(ArgumentError)
    end

    it 'makes the mcollective result object available after initialization' do
      Emcee::Result.new(@result).result.should == @result
    end

    it 'makes the result data available as raw contents after initialization' do
      Emcee::Result.new(@result).raw_contents.should == @result.results[:data]
    end

    it 'makes the result data available as an empty hash when empty result data was provided' do
      @results.delete(:data)
      Emcee::Result.new(@result).raw_contents.should == {}
    end

    it 'makes the node name from the result :sender available as node_name after initialization' do
      Emcee::Result.new(@result).node_name.should == @node_name
    end

    it 'makes the result status available as an integer status_code after initialization' do
      Emcee::Result.new(@result).status_code.should == @status_code
    end

    it 'makes the result status message available as a status message' do
      Emcee::Result.new(@result).status_message.should == @status_message
    end
  end

  describe 'when returning usable contents' do
    it 'returns the :properties value if it is set' do
      @contents[:properties] = { 'bar' => 'baz' }
      Emcee::Result.new(@result).contents.should == @contents[:properties]
    end

    it 'returns the :output value if no :properties are set but :output is set' do
      Emcee::Result.new(@result).contents.should == @contents[:output]
    end

    it 'returns the raw contents if neither :properties nor :output is set' do
      @contents.delete(:output)
      @contents['xyzzy'] = 'quux'
      Emcee::Result.new(@result).contents.should == @contents
    end
  end

  describe 'when determining whether a result was successful' do
    it 'returns true if the result status code is 0' do
      Emcee::Result.new(@result).successful?.should be_true
    end

    it 'returns false if the result status code is not 0' do
      @results[:statuscode] = 1
      Emcee::Result.new(@result).successful?.should be_false
    end
  end

  describe 'when returning result as a hash' do
    before do
      @result_data = File.read(fixture_location('single_puppetral_result.yml'))
      @result = YAML.load(@result_data)
    end

    it 'should represent the node_name value as a :node_name hash value' do
      @result.to_hash[:node_name].should == @result.node_name
    end

    it 'should represent the contents as a :contents hash value' do
      @result.to_hash[:contents].should == @result.contents
    end
  end
end

