require 'ostruct'
require 'csv_export'

class NodeReportController < ApplicationController
  unloadable
  helper :compliance

  include CSVExport

  def show
    _show('differences')
  end

  def failures
    _show('failures')
  end

  def all
    _show('all')
  end

  def export
    node = Node.find(params[:id])
    date = params[:date].present? ? Date.parse(params[:date]) : Time.zone.now.to_date
    node_baselines = NodeBaseline.self_comparisons
    node_baseline = node_baselines.find_by_node_id_and_date(params[:id], date,
      :include => {
        :baseline => {
          :resources => :properties,
        },
        :differences => {
          :baseline_resource => :properties,
          :report_resource => :events,
        },
      }
    )

    respond_to do |format|
      format.csv  do
        if node_baseline
          data = Diff.collate_properties_by_resource(node_baseline).map do |res, props|
            type, title = res
            props.map do |prop, vals|
              [
                node.name,
                node.name,
                type,
                title,
                prop,
                vals[:baseline] == vals[:inspected],
                vals[:baseline],
                vals[:inspected],
              ]
            end
          end
          data.flatten!(1)
        end
        send_csv :data => data, :filename => "#{date.to_s(:db)}-node-#{params[:id]}-custom.csv"
      end
    end
  end

  def list
    render 'shared/node_list'
  end

  def mark_as_reviewed
    diff = BaselineDifference.find(params[:diff_id])
    if state = review([diff])
      render :text => diff.node_baseline.resource_counts_by_status.to_json, :content_type => :json, :status => 200
    end
  end

  def mark_all_as_reviewed
    baseline = NodeBaseline.self_comparisons.find_by_node_id_and_date(params[:id], params[:date])
    diffs = baseline.differences
    if state = review(diffs)
      render :text => baseline.resource_counts_by_status.to_json, :content_type => :json, :status => 200
    end
  end

  def all_properties_for
    data, resource = nil, nil

    if diff = BaselineDifference.find(params[:diff_id])
      baseline = diff.baseline_resource

      if resource = diff.report_resource
        diff = Diff.node_diff_one(diff)

        # The full set of properties is:
        properties = resource.events.map(&:property) | baseline.properties.map(&:name)

        # property, baseline, report
        data = properties.reject {|x| diff[:properties].has_key? x }.map do |name|
          [name,
           baseline.properties.find_by_name(name).value,
           resource.events.find_by_property(name).previous_value
          ]
        end.sort_by {|x| x[0] }
      end
    end

    render(:partial => 'shared/one_resource_audited_only',
           :locals => { :data => data, :resource => resource })
  end

  private
  def review(changes)
    state = params[:state]

    unless ['accepted', 'rejected'].include?(state)
      render :text => "'#{state}' is not an acceptable state for a BaselineDifference", :status => 400
      return false
    end

    if state == 'accepted' then
      BaselineDifference.accept_all_changes(changes)
    else
      BaselineDifference.reject_all_changes(changes)
    end

    return state
  end

  def _show(tab)
    @tab = tab

    @node = Node.find(params[:id])

    @date = params[:date].present? ? Date.parse(params[:date]) : Time.zone.now.to_date

    @node_baseline = NodeBaseline.self_comparisons.
      find_by_node_id_and_date(params[:id], @date)

    @failure_count = ResourceStatus.all(
      :select => 'COUNT(1) AS count',
      :joins => :report,
      :group => 'report_id HAVING MAX(reports.time)',
      :conditions => {
        :failed => true,
        :reports => {
          :node_id => @node.id,
          :kind => 'inspect',
          :time => (@date...@date.tomorrow)
        }
      }
    ).map(&:count).map(&:to_i).sum

    if @node_baseline
      case tab
      when 'differences' then
        params[:page] ||= 1
        params[:per_page] ||= BaselineDifference.per_page
        total_count = @node_baseline.differences.count
        @baseline_differences = WillPaginate::Collection.create(params[:page], params[:per_page], total_count) do |pager|
          pager.replace Diff.node_against_baseline(params[:id], @date, pager.offset, pager.per_page)
        end

      when 'failures' then
        redirect_to(:action => :show, :date => @date.to_s(:db)) and return if @failure_count.zero?

        params[:page] ||= 1
        params[:per_page] ||= BaselineDifference.per_page

        @failures = WillPaginate::Collection.create(params[:page], params[:per_page], @failure_count) do |pager|
          inspection = @node.reports.find(:first,
            :order      => 'time DESC',
            :conditions => {
              :kind => 'inspect',
              :time => (@date...@date.tomorrow),
            }
          )

          resources = inspection.resource_statuses.all(
            :conditions => { :failed => true },
            :offset     => pager.offset,
            :limit      => pager.per_page
          )

          baseline_resources = @node_baseline.baseline.resources.where([
            (["(resource_type = ? AND title = ?)"] * resources.length).join(' OR '),
            *resources.map { |e| [e.resource_type, e.title] }.flatten
          ]).group_by(&:name)

          data = resources.map do |e|
            res = (baseline_resources[e.name] || []).first
            [
              e.name,
              (res ? res.properties.map { |p| [p.name, p.value] } : [['',' ']]),
              inspection.logs.map(&:message).grep(Regexp.new(Regexp.escape e.name)),
            ]
          end

          pager.replace data
        end

      when 'all' then
        time = BaselineUtil.day_beginning(@date)..BaselineUtil.day_ending(@date)
        report = Report.find(:first, :conditions => {
                               :kind    => 'inspect',
                               :time    => time,
                               :node_id => @node_baseline.node
                             },
                             :order => 'time DESC')
        if report then
          statuses = report.resource_statuses.all(:order => "resource_type, title")
          @resources = paginate_scope(statuses)
        end
      else
        raise "Unknown tab #{tab} found!"
      end

      @status_counts = @node_baseline.resource_counts_by_status
    elsif tab == 'failures'
      redirect_to(:action => :show, :date => @date.to_s(:db)) and return
    end

    @unreviewed_changes_on_other_days = BaselineUtil.unreviewed_changes_on_other_days(@date, @node.id)
    @other_days = BaselineUtil.date_changer_options(@date, @node.id)

    render 'node/show'
  end
end
